# FILE: opendlg.tcl
#
# Open file dialog box for mmDisp
#
# Callbacks: OPEN
#            DELETE $delete_callback_arg
Oc_Class OpenDlg {
    const public variable winpath
    const public variable outer_frame_options = "-bd 2 -relief raised"
    const public variable filter = "*.o?f *.svf *.vio"
    const public variable dialog_title = "Open File"
    public variable callback = Oc_Nop
    public variable delete_callback_arg = {}
    private variable filewidget = {}
    private variable autobox = {}
    private variable ctrlwidget = {}
    private variable selectionwidget = {}
    private variable dirwidget = {}
    Constructor {args} {
	eval $this Configure -winpath .w$this $args
	# Create toplevel frame to hold dialog
	toplevel $winpath
	wm group $winpath .
	wm title $winpath $dialog_title
	if {![string match {} [info commands Ow_PositionChild]]} {
	    Ow_PositionChild $winpath  ;# Position at +.25+.25 over '.'
	}
	wm protocol $winpath WM_DELETE_WINDOW \
		"$this SwitchBoard $this CANCEL"
      focus $winpath

	# Temporarily disable callbacks until construction is complete
	set callback_keep $callback ;	set callback Oc_Nop

	# Create subwidgets
	Ow_FileSelect New filewidget $winpath \
		-callback "$this SwitchBoard" \
		-filter $filter
	set ctrlframe [frame $winpath.ctrlframe]
	AutoBox New autobox $winpath
	## Note: The AutoBox class is defined below
	CtrlWBrowse New ctrlwidget $winpath -callback "$this SwitchBoard"
	Ow_SelectBox New selectionwidget $winpath \
		-callback "$this SwitchBoard"
	Ow_DirSelect New dirwidget $winpath -callback "$this SwitchBoard"

	# Initialize selection path
	$selectionwidget SetPath [$dirwidget Cget -value]

	# Pack subwidgets
	pack [$filewidget Cget -winpath] -side left \
		-fill both -expand 1 -padx 2m -pady 2m
	pack [$ctrlwidget Cget -winpath] -in $ctrlframe \
		-side right -expand 0 -anchor ne -padx 2m -pady 2m
	pack [$autobox Cget -winpath] -in $ctrlframe \
		-side top -expand 0 -padx 2m -pady 2m
	pack $ctrlframe -side top \
		-fill x -expand 0 -anchor ne -padx 2m -pady 2m
	pack [$selectionwidget Cget -winpath] -side top \
		-fill x -expand 0 -anchor n -padx 2m -pady 2m
	pack [$dirwidget Cget -winpath] -side right \
		-fill both -expand 1 -padx 2m -pady 2m

	# Widget bindings
	bind $winpath <Destroy> "+$this WinDestroy %W"

	# Re-enable callbacks
	set callback $callback_keep
    }
    method GetOptionState {} { autobox } {
	# Returns a list of autobox state pairs,
	# e.g., -autosample 0 -autozoom 1
	return [$autobox GetState]
    }
    method GetFilename {} { selectionwidget } {
	# Returns full path+filename of selected value
	return [$selectionwidget ReadDisplayedFullname]
    }
    method SwitchBoard { subwidget actionid args } \
	    { callback ctrlwidget filewidget dirwidget selectionwidget } {
	# General callback routing routine

	switch -exact -- $actionid {
	    UPDATE {
		# Update actions
		if {[string compare $subwidget $filewidget]==0} {
		    $selectionwidget SetFilename [lindex $args 1]
		} elseif {[string compare $subwidget $dirwidget]==0} {
		    $filewidget SetPath [lindex $args 1]
		    $selectionwidget SetPath [lindex $args 1]
		} else {
		    error "Unknown UPDATE callback source: $subwidget"
		}
	    }
	    SELECT {
		set errmsg [eval $callback $this OPEN]
		if { ![$ctrlwidget Cget -browse_state] \
			&& [string match {} $errmsg] } {
		    $this Delete ; return
		}
	    }
	    CANCEL {
		$this Delete ; return
	    }
	    RESCAN {
		$filewidget Rescan
		$dirwidget Rescan
	    }
	    DELETE {}
	    default { error "Unknown actionid: $actionid" }
	}
    }
    method WinDestroy { w } { winpath } {
	if {[info exists winpath] && [string compare $winpath $w]!=0} {
	    return    ;# Child destroy event
	}
	$this Delete
    }
    private variable delete_in_progress = 0
    Destructor {
	# Protect against re-entrancy
	if { [info exists delete_in_progress] && \
		$delete_in_progress } { return }
	set delete_in_progress 1

	# Notify client of impending destruction
	catch {eval $callback $this DELETE {$delete_callback_arg}}

	# Delete subwidgets
	catch {$filewidget      Delete}
	catch {$autobox         Delete}
	catch {$ctrlwidget      Delete}
	catch {$selectionwidget Delete}
	catch {$dirwidget       Delete}

	# Delete window
	if {[info exists winpath] && [winfo exists $winpath]} {
	    # Destroy instance frame, all children, and bindings
	    destroy $winpath
	}
    }
}

# Open parameter "auto" options box for mmDisp File|Open dialog
#
# NOTE: This subwidget does not make callbacks.
;Oc_Class AutoBox {
    const public variable winpath
    const public variable outer_frame_options = "-bd 4 -relief ridge"
    public variable autosample = 1
    public variable autozoom = 1
    Constructor {basewinpath args} {
	if {![string match {.} \
		[string index $basewinpath \
		[expr [string length $basewinpath]-1]]]} {
	    append basewinpath {.}
	}
	eval $this Configure -winpath ${basewinpath}w$this $args

	# Pack entire subwidget into a subframe.
	eval frame $winpath $outer_frame_options
	set innerframe [frame $winpath.innerframe -relief raised -bd 2]
	set checkframe [frame $winpath.checkframe -relief sunken -bd 2]

	set autolabel [label $winpath.label -text "Auto:" \
		-relief flat -padx 4 -pady 4]

	set but1 [checkbutton $winpath.btn1 -text "Subsample" \
		-relief flat -anchor w \
		-variable [$this GlobalName autosample]]
	set but2 [checkbutton $winpath.btn2 -text "Zoom" \
		-relief flat -anchor w \
		-variable [$this GlobalName autozoom]]

	pack $innerframe
	pack $autolabel $checkframe -in $innerframe \
		-side left -anchor ne
	pack $but1 $but2 -in $checkframe \
		-side top -expand 1 -fill both

	# Widget bindings
	bind $winpath <Destroy> "$this WinDestroy %W"
    }
    method GetState {} { autosample autozoom } {
	set statelist {}
	foreach i { autosample autozoom } {
	    lappend statelist -$i [set $i]
	}
	return $statelist
    }
    method WinDestroy { w } { winpath } {
	if {[info exists winpath] && [string compare $winpath $w]!=0} {
	    return    ;# Child destroy event
	}
	$this Delete
    }
    Destructor {
	if {[info exists winpath] && [winfo exists $winpath]} {
            # Remove <Destroy> binding
	    bind $winpath <Destroy> {}
	    # Destroy instance frame, all children, and bindings
	    destroy $winpath
	}
    }
}
