# FILE: dialog.tk
#
# Tcl/Tk scripts for various dialog boxes in mmGraph
#

###################### Plot Select and Configuration #####################
# 25-June-1998, mjd
#
# Local configuration results are stored in array psc.
Oc_Class PlotConfigure {
    const public variable title = "mmGraph: Configure"
    const public variable winpath

    # Import trace and initialization array.  Must be global.
    const public variable import_arrname = {}

    # Routine in mainline to call on an Apply or OK event.
    public variable apply_callback = Oc_Nop 

    # Menu and index (as a 2-item list) in the launching window
    # to toggle off in the constructor and on in the destructor.
    const public variable menu = {}

    # Internal state array
    private array variable psc

    # State (normal/disabled) dependency lists
    private array variable dep

    Constructor {args} {
        # Create a (presumably) unique name for the toplevel
        # window based on the instance name.
        set winpath ".[$this WinBasename]"

	# Process user preferences
	eval $this Configure $args

        # Create dialog window
        toplevel $winpath
        wm group $winpath .
        wm title $winpath $title
        Ow_PositionChild $winpath ;# Position at +.25+/25 over '.'
        focus $winpath

        # Setup destroy binding.
	bind $winpath <Destroy> "$this WinDestroy %W"

        # Disable launch menu, if given
        if {![string match {} $menu]} {
            [lindex $menu 0] entryconfigure [lindex $menu 1] \
                    -state disabled
        }

        # Setup internal state array default values
        array set psc {
	    title  {}
	    autolabel 1
	    xlabel {}
	    ylabel {}
	    autolimits 1
	    xmin   0
	    xmax   1
	    ymin   0
	    ymax   1
            ptbufsize  0
	}
        set gpsc [$this GlobalName psc]
        ## Global name of psc array.  This is needed to tie
        ## elements of the array to Tk buttons.

	# Determine "disabled" color
	checkbutton $winpath.cktemp
        set disabledcolor [$winpath.cktemp cget -disabledforeground]
	destroy $winpath.cktemp

        # Allow override from user initialization array
        if {![string match {} $import_arrname]} {
            upvar #0 $import_arrname userinit
            if {[info exists userinit]} {
                foreach elt [array names userinit] {
                    if {[info exists psc($elt)]} {
                        set psc($elt) $userinit($elt)
                    }
                }
            }
            trace variable userinit w "$this ExternalUpdate $import_arrname"
        }

	# Title
        Ow_EntryBox New _ $winpath \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {Title:} -valuewidth 0 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(title)
	pack [$_ Cget -winpath] -fill x -expand 1 -padx 10 -pady 10

	# Frame for interior grid
	set gf [frame $winpath.gf]
	
	# Labels
	set dep(autolabel) {}
        checkbutton $gf.autolabel -text "Auto Label" \
                -variable ${gpsc}(autolabel) \
                -onvalue 1 -offvalue 0 \
		-command "$this UpdateState autolabel"

	label $gf.xlabel -text "X Label:"
        Ow_EntryBox New _ $gf \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 0 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(xlabel)
	lappend dep(autolabel) $_
	grid $gf.autolabel $gf.xlabel [$_ Cget -winpath]
	grid $gf.autolabel -sticky nw -padx 10
	grid $gf.xlabel -sticky w
	grid [$_ Cget -winpath] -sticky ew

	label $gf.ylabel -text "Y Label:"
        Ow_EntryBox New _ $gf \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 0 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(ylabel)
	lappend dep(autolabel) $_
	grid ^ $gf.ylabel [$_ Cget -winpath]
	grid $gf.ylabel -sticky w
	grid [$_ Cget -winpath] -sticky ew

        $this UpdateState autolabel

	# Spacer
	set rowcnt [lindex [grid size $gf] 1]
	grid rowconfigure $gf $rowcnt -minsize 10

	# Scaling
	set dep(autolimits) {}
        checkbutton $gf.autolimits -text "Auto Scale" \
                -variable ${gpsc}(autolimits) \
                -onvalue 1 -offvalue 0 \
		-command "$this UpdateState autolimits"

	label $gf.xmin -text "X Min:"
        Ow_EntryBox New _ $gf \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 0 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(xmin) -valuetype float
	lappend dep(autolimits) $_
	set rowcnt [lindex [grid size $gf] 1]
	grid $gf.autolimits $gf.xmin [$_ Cget -winpath] -row $rowcnt
	grid $gf.autolimits -sticky nw -padx 10
	grid $gf.xmin -sticky w
	grid [$_ Cget -winpath] -sticky ew

	label $gf.xmax -text "X Max:"
        Ow_EntryBox New _ $gf \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 0 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(xmax) -valuetype float
	lappend dep(autolimits) $_
	grid ^ $gf.xmax [$_ Cget -winpath]
	grid $gf.xmax -sticky w
	grid [$_ Cget -winpath] -sticky ew

	label $gf.ymin -text "Y Min:"
        Ow_EntryBox New _ $gf \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 0 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(ymin) -valuetype float
	lappend dep(autolimits) $_
	grid ^ $gf.ymin [$_ Cget -winpath]
	grid $gf.ymin -sticky w
	grid [$_ Cget -winpath] -sticky ew

	label $gf.ymax -text "Y Max:"
        Ow_EntryBox New _ $gf \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 0 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(ymax) -valuetype float
	lappend dep(autolimits) $_
	grid ^ $gf.ymax [$_ Cget -winpath]
	grid $gf.ymax -sticky w
	grid [$_ Cget -winpath] -sticky ew

	foreach {colcount rowcount} [grid size $gf] {}
	grid columnconfigure $gf [expr $colcount-1] -weight 1

	$this UpdateState autolabel
	$this UpdateState autolimits
	pack $gf -fill x -expand 1 -padx 10

        # Point buffer
        Ow_EntryBox New _ $winpath \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {Point buffer size:} -valuewidth 6 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(ptbufsize) -valuetype posint
	pack [$_ Cget -winpath] -fill none -expand 0 -anchor w \
                -padx 10 -pady 10


        # Control buttons
	set ctrlframe [frame $winpath.ctrlframe]

        set cancel [button $ctrlframe.cancel -text "Cancel" \
                -command "$this Action cancel" ]
        set apply [button $ctrlframe.apply -text "Apply" \
                -command "$this Action apply" ]
        set ok [button $ctrlframe.ok -text "OK" \
                -command "$this Action ok" ]
        pack $cancel $apply $ok -side left -expand 1 -padx 20

	pack $ctrlframe -fill x -expand 1 -padx 10 -pady 10

        wm protocol $winpath WM_DELETE_WINDOW "$cancel invoke"
	Ow_SetIcon $winpath
    }

    # Update from external source
    callback method ExternalUpdate { globalarrname name elt ops } {
        ## The trace is setup to pass in the global array name so that
        ## even if the trace is triggered by a call to Tcl_SetVar
        ## in C code called from a proc from which the traced array
        ## is not visible (due to no 'global' statement on that array),
        ## this trace can still function.        
        if {[string match {} $elt]} {return}
        upvar #0 $globalarrname source
        if {[info exists psc($elt)] && \
                ![string match $psc($elt) $source($elt)]} {
            set psc($elt) $source($elt)
        }
    }

    # Plot type status change helper function
    callback method UpdateState { type } {
	# Note: "type" here should be either autolabel
	#  or autolimits.  In either case, if the "auto"
	#  button is on, then the dependency widgets
	#  should be _disabled_.
        if {$psc($type)} {
            set mystate disabled
        } else {
            set mystate normal
        }
        foreach win $dep($type) {
            $win configure -state $mystate
        }
    }

    # Dialog action routine
    callback method Action { cmd } {
        switch $cmd {
            cancel {
                $this Delete
                return
            }
            apply {
                eval $apply_callback psc
            }
            ok {
                eval $apply_callback psc
                $this Delete
                return
            }
            default {
                error "Illegal action code: $cmd"
            }
        }
    }

    # Destructor methods
    method WinDestroy { w } { winpath } {
	if {[string compare $winpath $w]!=0} { return } ;# Child destroy event
	$this Delete
    }
    Destructor {
        # Re-enable launching menu
        if {![string match {} $menu]} {
            set menuwin [lindex $menu 0]
            if {[winfo exists $menuwin]
                    && [catch { $menuwin entryconfigure \
                    [lindex $menu 1] -state normal } errmsg]} {
                Oc_Log Log $errmsg error
            }
        }
        # Remove external update trace, if any
        if {[catch {
            if {![string match {} $import_arrname]} {
                upvar #0 $import_arrname userinit
                trace vdelete userinit w "$this ExternalUpdate $import_arrname"
            }
        } errmsg]} {
            Oc_Log Log $errmsg error
        }
        # Destroy windows
	if {[catch { 
            if {[info exists winpath] && [winfo exists $winpath]} {
                # Remove <Destroy> binding
                bind $winpath <Destroy> {}
                # Destroy toplevel, all children, and all bindings
                destroy $winpath
            }
        } errmsg]} {
            Oc_Log Log $errmsg error
        }
    }
}
