#!/bin/sh
# FILE: gif2ppm.tcl
#
# NOTE: This file's name is a misnomer.  This script will convert
# data in any image format understood by Tk's (possibly
# extended) [image photo] command to data representing the same
# image in the PPM P3 format.  It knows more than just GIF.
#
# This comment continues on in Tcl, but not in sh \
exec wish "$0" ${1+"$@"}
#
# If the above exec doesn't work, replace ${1+"$@"} with "$@"
# The syntax ${1+"$@"} means: if $1 (positional parameter #1)
# is set, then substitute "$@" (all parameters as separate words),
# if $1 is not set (no command line arguments), then substitute
# nothing.  Bash makes no substitution for "$@" if $1 is not set,
# so these are equivalent under bash.  Apparently there is some
# danger that a shell may substitute "" for "$@" if $1 is not set,
# but I don't know which shells do that.

if {$argc<1 || [string match "-h*" [lindex $argv 0]]} {
    puts stderr {Usage: gif2ppm [-noinfo]\
	    [-o <{}|outfile|->] file1 [file2 ...]}
    exit 1
}

set noinfo 0
set flag [lsearch -regexp $argv {^--?no_?info$}]
if {$flag>=0} {
    set noinfo 1
    set argv [lreplace $argv $flag $flag]
}

set errcount 0

set BUFSIZ 250000       ;# Size of output write buffer

# Load Tk extension for windows: Under Windows, wish has no standard
#  channels, so to use this script as a filter, it must be launched with
#  tclsh.  But we need the image and photo code in Tk.
# This is damned ugly! -- it ought to be a simple [package require TK]
if {[string match windows $tcl_platform(platform)]} {
    set tclver [join [split [package provide Tcl] .] {}]
    if {$tclver < 80} {
        set tkver [expr {$tclver - 34}]
    } else {
        set tkver $tclver
    }
    if {[info exists env(TERM)] 
	    && [string match cygwin [string tolower $env(TERM)]]} {
	# Using cygwin tool set, so name of Tk dll is cygtk*.dll
	catch {load cygtk$tkver[info sharedlibextension] Tk}
    } else {
	catch {load tk$tkver[info sharedlibextension] Tk}
    }
}

proc make_ppm_name { inname } {
    set outname [file rootname $inname].ppm
    if {[file exists $outname]} {
	set basename $outname
	set outname {}
	for {set i 1} {$i<100} {incr i} {
	    set testname ${basename}.$i
	    if {![file exists $testname]} {
		set outname $testname
		break
	    }
	}
    }
    return $outname
}

proc write_ppm_file { chanid pic } {
    # The Tk photo interface does not have a "write to channel" option.
    set width [image width $pic] ; set height [image height $pic]
    puts $chanid "P3"
    puts $chanid "$width $height"
    puts $chanid "255"
    set rtop [expr $height-1] ; set ctop [expr $width-1]
    for {set r 0} {$r<$height} {incr r} {
	for {set c 0} {$c<$width} {incr c} {
	    puts $chanid [$pic get $c $r]
	}
    }
    flush $chanid
}

set outloc name   ;# One of name|file|stdout
set outchan {}

for {set i 0} {$i<[llength $argv]} {incr i} {
    set inname [lindex $argv $i]
    if {[string match "-o" $inname]} {
	# Output redirection request
	set outname [lindex $argv [incr i]]
	switch -exact -- $outname {
	    {} {
		set outloc name
	    }
	    {-} {
		if {[string match file $outloc]} { close $outchan }
		set outloc stdout
		set outchan stdout
	    }
	    default {
		set outloc file
		if {[catch {open $outname "w"} outchan]} {
		    set outloc name
		    puts stderr "Unable to open output file $outname."
		    incr errcount
		}
                fconfigure $outchan -translation binary \
                        -buffering full -buffersize $BUFSIZ
	    }
	}
	continue
    }
    # Open input file.
    if {![file readable $inname]} {
	puts stderr "Unable to open input file $inname; Skipping."
	incr errcount
	continue
    }
    if {!$noinfo} { puts stderr "Processing input file $inname." }
    if {[catch {set pic [image create photo -file $inname]} errmsg]} {
	puts stderr "ERROR: $errmsg"
	incr errcount
	continue
    }
    if {!$noinfo} { puts stderr "Processing output..." }
    # Open output file, if necessary.
    if {[string match name $outloc]} {
	set outname [make_ppm_name $inname]
	if {[string match {} $outname]} {
	    puts stderr "Unable to generate unique output name; Skipping."
	    incr errcount
	    continue
	}
	if {[catch {open $outname "w"} outchan]} {
	    puts stderr "Unable to open output file $outname; Skipping."
	    incr errcount
	    continue	    
	}
        fconfigure $outchan -translation binary \
                -buffering full -buffersize $BUFSIZ
    }

    if {[catch {write_ppm_file $outchan $pic} errmsg]} {    
	puts stderr "PPM WRITE ERROR: $errmsg"
	incr errcount
    }
    if {!$noinfo} { puts stderr "Output written to $outname" }
    if {[string match name $outloc]} { close $outchan }
}

exit $errcount
