/* FILE: maganis.cc	      -*-Mode: c++-*-
 *
 * This file contains function definitions for the mag anisotropy
 *  initialization routines.
 *
 * NOTE: These functions directly access the Grid2D's magnetization
 *       array, which use the xzy coordinate system.  For efficiency,
 *       these routines also use those coordinates, and the anisotropy
 *       directions passed in to these functions are assumed to be
 *       in the xzy coordinate system.  HOWEVER, the label identifiers
 *       for these functions in MIF files need to use the user xyz
 *       coordinates, so there may be an apparent mismatch between
 *       function and label names.  For example, the UniformXZAnisInit
 *       routine uses a MIF file label of "uniformxy."
 *
 */


#include <math.h>

#include "maganis.h"
#include "vecio.h"
#include "vector.h"

/* End includes */

///////////////////////////////////////////////////////////////////////
/// MagAnisInitListNode
///
void MagAnisInitListNode::AddNode(MagAnisInitListNode* new_node)
{
  if(next_node==NULL) next_node=new_node;
  else                next_node->AddNode(new_node);
}

MagAnisInitListNode::MagAnisInitListNode(ANIS_INIT_CHECK_FUNC* classcheck,
					 MagAnisInitListNode* firstnode)
{
  anischeck=classcheck;
  next_node=NULL;
  if(firstnode!=NULL) {
    firstnode->AddNode(this);
  }
}

ANIS_INIT_FUNC* MagAnisInitListNode::GetAnisInitFunc(const char* name)
{ // Returns NULL if no match found
  ANIS_INIT_FUNC* aifp=anischeck(name);
  if(aifp!=NULL || next_node==NULL) return aifp;
  return next_node->GetAnisInitFunc(name);
}


///////////////////////////////////////////////////////////////////////
/// Static (crystalline) anisotropy initialization classes
///////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////
/// Dummy first element
///
static ANIS_INIT_FUNC* DummyTypeCheck(const char*) { return NULL; }
MagAnisInitListNode AnisInitList(DummyTypeCheck,NULL);

///////////////////////////////////////////////////////////////////////
/// Constant (in space) anisotropy orientation
///
static ANIS_INIT_FUNC ConstantAnisInit;
static ANIS_INIT_CHECK_FUNC ConstantAnisTypeCheck;
static MagAnisInitListNode ConstantAnis(ConstantAnisTypeCheck,&AnisInitList);

ANIS_INIT_FUNC* ConstantAnisTypeCheck(const char* name)
{
  if(strcmp(name,"constant")==0) return ConstantAnisInit;
  return NULL;
}

int ConstantAnisInit(/* Imports: */
		     int /* argc */,const char* const* /* argv */,
		     int Nx,int Nz,MagEltAnisType meat,
		     double K1,ThreeVector& dirA,
		     ThreeVector& dirB,ThreeVector& dirC,
		     /* Exports: */
		     MagElt** m,ThreeVector** &anisA,
		     ThreeVector** &anisB,ThreeVector** &anisC)
{
  int i,k;
  // Initialize K1 inside each magelt.
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m[i][k].SetK1(K1);
  }

  // Setup anisotropy direction arrays (1 or 0 elements!)
  double *aptr=NULL,*bptr=NULL,*cptr=NULL;
  ThreeVector::Alloc2DArray(1,1,anisA);
  anisA[0][0]=dirA; /* Copy data */
  aptr=anisA[0][0];
  if(meat==meat_uniaxial) { anisB=anisC=NULL; }
  else {
    ThreeVector::Alloc2DArray(1,1,anisB);
    anisB[0][0]=dirB; /* Copy data */
    bptr=anisB[0][0];
    ThreeVector::Alloc2DArray(1,1,anisC);
    anisC[0][0]=dirC; /* Copy data */
    cptr=anisC[0][0];
  }

  // Point MagElt anisdir pointers into above arrays
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m[i][k].InitAnisDirs(aptr,bptr,cptr);
  }

  return 0;
}

///////////////////////////////////////////////////////////////////////
/// Uniformly random in XZ-plane anisotropy orientation (i.e., $K\in S^1$)
///
static ANIS_INIT_FUNC UniformXZAnisInit;
static ANIS_INIT_CHECK_FUNC UniformXZAnisTypeCheck;
static MagAnisInitListNode UniformXZAnis(UniformXZAnisTypeCheck,&AnisInitList);

ANIS_INIT_FUNC* UniformXZAnisTypeCheck(const char* name)
{
  // NOTE: The apparent mismatch between the MIF label name (uniformxy)
  //  and these routine names (UniformXZ) is due to the the use of
  //  xzy coords internally, and xyz coords externally.  See the NOTE
  //  at the top of this file.
  if(strcmp(name,"uniformxy")==0) return UniformXZAnisInit;
  return NULL;
}

int UniformXZAnisInit(/* Imports: */
		     int /* argc */,const char* const* /* argv */,
		     int Nx,int Nz,MagEltAnisType meat,
		     double K1,ThreeVector& /* dirA */,
		     ThreeVector& /* dirB */,ThreeVector& /* dirC */,
		     /* Exports: */
		     MagElt** m,ThreeVector** &anisA,
		     ThreeVector** &anisB,ThreeVector** &anisC)
{
  int i,k;
  int anis_count=3;
  if(meat==meat_uniaxial) anis_count=1;

  // Initialize K1 inside each magelt.
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m[i][k].SetK1(K1);
  }

  // Allocate anis array(s)
  ThreeVector::Alloc2DArray(Nx,Nz,anisA);
  if(anis_count==1) { anisB=anisC=NULL; }
  else {
    ThreeVector::Alloc2DArray(Nx,Nz,anisB);
    ThreeVector::Alloc2DArray(Nx,Nz,anisC);
  }

  // Initialize anis arrays
  double theta,x,z;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    theta=UnifRand()*2*PI;
    x=cos(theta);
    z=1-x*x;
    if(z<0.0) z=0.0; else z=sqrt(z); // Safety
    if(theta>PI) z*=-1;
    anisA[i][k].Set(x,0.,z);
    if(anis_count!=1) {
      anisB[i][k].Set(-z,0.,x);
      anisC[i][k].Set(0.,1.,0.);
    }
  }

  // Point MagElt anisdir pointers into above anis arrays
  if(anis_count==1) {
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
      m[i][k].InitAnisDirs(anisA[i][k],NULL,NULL);
  }
  else {
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
      m[i][k].InitAnisDirs(anisA[i][k],anisB[i][k],anisC[i][k]);
  }

  return 0;
}

///////////////////////////////////////////////////////////////////////
/// Uniformly random in 3-space anisotropy orientation (i.e., $K\in S^2$)
///
static ANIS_INIT_FUNC UniformS2AnisInit;
static ANIS_INIT_CHECK_FUNC UniformS2AnisTypeCheck;
static MagAnisInitListNode UniformS2Anis(UniformS2AnisTypeCheck,
					 &AnisInitList);

ANIS_INIT_FUNC* UniformS2AnisTypeCheck(const char* name)
{
  if(strcmp(name,"uniforms2")==0) return UniformS2AnisInit;
  return NULL;
}

int UniformS2AnisInit(/* Imports: */
		     int /* argc */,const char* const* /* argv */,
		     int Nx,int Nz,MagEltAnisType meat,
		     double K1,ThreeVector& /* dirA */,
		     ThreeVector& /* dirB */,ThreeVector& /* dirC */,
		     /* Exports: */
		     MagElt** m,ThreeVector** &anisA,
		     ThreeVector** &anisB,ThreeVector** &anisC)
{
  int i,k;
  int anis_count=3;
  if(meat==meat_uniaxial) anis_count=1;

  // Initialize K1 inside each magelt.
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m[i][k].SetK1(K1);
  }

  // Allocate anis array(s)
  ThreeVector::Alloc2DArray(Nx,Nz,anisA);
  if(anis_count==1) { anisB=anisC=NULL; }
  else {
    ThreeVector::Alloc2DArray(Nx,Nz,anisB);
    ThreeVector::Alloc2DArray(Nx,Nz,anisC);
  }

  // Initialize anis arrays
  double dummy1arr[3],dummy2arr[3],dummy3arr[3];
  ThreeVector dummy1(dummy1arr),dummy2(dummy2arr),dummy3(dummy3arr);
  double dot;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    dummy1.Random(1.0);  // Get random vector
    anisA[i][k]=dummy1;
    if(anis_count!=1) {
      do {
	dummy3.Random(1.0);
	dot=dummy1*dummy3;
      } while(fabs(dot)>0.999);
      dummy2.FastAdd(dummy3,-dot,dummy1);
      dummy2.Scale(1.0);
      anisB[i][k]=dummy2;
      dummy2^=dummy1;
      dummy2.Scale(1.0);  // Safety
      anisC[i][k]=dummy2;
    }
  }

  // Point MagElt anisdir pointers into above anis arrays
  if(anis_count==1) {
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
      m[i][k].InitAnisDirs(anisA[i][k],NULL,NULL);
  }
  else {
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
      m[i][k].InitAnisDirs(anisA[i][k],anisB[i][k],anisC[i][k]);
  }

  return 0;
}
