# File: hystctrl.tcl
#
# Sample micromagnetic control script.  Must be run using the
# mmsolve shell.
#
if {$argc>0} { set mif_file [lindex $argv 0]} else {
    puts stderr "Usage: mmsolve hystctrl.tcl <mif-file>"
    exit 1
}

# Support libraries
package require Oc
package require Tk
package require Ow

# Initialize solver
mms_mif New mif
$mif Read $mif_file
mms_solver New solver $mif
set Bfield_multiplier 1000  ;# Convert T to mT

if {$argc>3} { set Bdir_nom [lrange $argv 1 3] }
if {$argc>4} { set Bdir_label [lindex $argv 4] }
## Bdir_label is label to put on M vs. B loop
##
## If Bdir_nom (nominal applied field direction) is left unset
## (i.e., not specified on the command line), then the code
## assumes the nominal direction is along one of the coordinate
## axis, and tries to extract it from the MIF file.

proc Dot { vec1 vec2 } {
    # Dot product of two 3-lists
    return    [expr [lindex $vec1 0]*[lindex $vec2 0] \
	    +       [lindex $vec1 1]*[lindex $vec2 1] \
	    +       [lindex $vec1 2]*[lindex $vec2 2]]
}

# Guess Bdir_nom, if necessary
if {![info exists Bdir_nom]} {
    set Bdir_nom [$solver GuessAppliedFieldAxis]
}
if {![info exists Bdir_label]} {
    if {[Dot $Bdir_nom "1 0 0"] > .9} {
	set Bdir_label "B_x applied (mT)"
    } elseif {[Dot $Bdir_nom "0 1 0"] > .9} {
	set Bdir_label "B_y applied (mT)"
    } elseif {[Dot $Bdir_nom "0 0 1"] > .9} {
	set Bdir_label "B_z applied (mT)"
    } else {
	set Bdir_label "B applied (mT)"
    }
}


set frame_opts "-bd 4 -relief ridge -background #0000cd"

# Text output subroutines
proc PrintState { solver } {
    set stepcount [$solver GetODEStepCount]
    set mxh [$solver GetMaxMxH]
    set maxangle [$solver GetMaxAngle]
    set energies [join [$solver GetEnergyDensities]]
    puts [format "%5d %10.7f  %5.1f  %12.4f %12.4f %12.4f %12.4f %12.4f" \
	    $stepcount $mxh $maxangle \
	    [lindex $energies 1] [lindex $energies 3] \
	    [lindex $energies 5] [lindex $energies 7] \
	    [lindex $energies 9]]
}

proc PrintLabels {} {
puts "  ODE     Max       Max                               ENERGIES"
puts " Step   |m x h|    Angle     Total       Exchange    Anisotropy     Demag        Zeeman"
}

# Output update subroutines
proc UpdateCurrentGraph {} {
    global current_graph
    set limits_change [$current_graph SetGraphLimits]
    if {$limits_change} {
        $current_graph RefreshDisplay
    } else {
        $current_graph DrawCurves
    }
}

proc UpdateOutput { {ordinate "iterate"} } {
    # Import $ordinate should be either "iterate" or "bapp"
    # This proc keeps track of the last "iterate" iteration
    # and "bapp" iteration processed, and does not repeat.
    global solver Bdir_nom Bfield_multiplier
    global graphTi graphMi graphEi graphAngi
    global graphTb graphMb graphEb graphAngb

    set iter_count [$solver GetODEStepCount]
    if {![string match "iterate" $ordinate]} {
	# Applied field update request
	global last_bapp_record_iter
	if {[info exists last_bapp_record_iter] && \
		$last_bapp_record_iter == $iter_count} {
	    return   ;# This field step already processed
	}
	set last_bapp_record_iter $iter_count
	set tg $graphTb   ; set mg $graphMb
	set eg $graphEb   ; set angg $graphAngb
	set xval [expr $Bfield_multiplier * \
		[Dot $Bdir_nom [$solver GetNomAppliedField]]]
	# Notify log of field stepping
	puts [format "### Field step %3d at iteration %5d" \
		[$solver GetFieldStepCount] $iter_count]
	# Dump magnetization
	set outname [format "%s.fs%04dm.vio" \
		[$solver GetOutputBasename] [$solver GetFieldStepCount]]
	set note [format "Relaxed magnetization state at iteration %5d" \
		$iter_count]
	$solver WriteMagFile $outname $note
    } else {
	global last_record_iter
	if {[info exists last_record_iter] && \
		$last_record_iter == $iter_count} {
	    return   ;# This field step already processed
	}
	set last_record_iter $iter_count
	# ODE iteration step request
	set tg $graphTi   ; set mg $graphMi
	set eg $graphEi   ; set angg $graphAngi
	set xval $iter_count
	# Dump iteration log
	if {$iter_count % 10 == 0} {
	    set mxh [$solver GetMaxMxH]
	    puts "Pass [format "%4d" $iter_count]: \
		    Max |m x h| =[format "%7.5f" $mxh]"
	}
    }

    set mxh [$solver GetMaxMxH]
    $tg AddDataPoints "|m x h|" $xval $mxh

    set mag [$solver GetAveMag]
    $mg AddDataPoints "Mx" $xval [lindex $mag 0]
    $mg AddDataPoints "My" $xval [lindex $mag 1]
    $mg AddDataPoints "Mz" $xval [lindex $mag 2]

    set energies [join [$solver GetEnergyDensities]]
    $eg AddDataPoints "Total\nEnergy"      $xval [lindex $energies 1]
    $eg AddDataPoints "Exchange\nEnergy"   $xval [lindex $energies 3]
    $eg AddDataPoints "Anisotropy\nEnergy" $xval [lindex $energies 5]
    $eg AddDataPoints "Demag\nEnergy"      $xval [lindex $energies 7]
    $eg AddDataPoints "Zeeman\nEnergy"     $xval [lindex $energies 9]

    set maxang [$solver GetMaxAngle]
    $angg AddDataPoints "Max\nAngle" $xval $maxang
    UpdateCurrentGraph
}

proc InspectLoop { solver } {
    after cancel InspectLoop $solver ;# There can be only 1. ;-/
    UpdateOutput iterate
    update idletasks  ;# Let the above output requests be
    ## handled before putting InspectLoop back into the
    ## event stack (otherwise if it takes too long to service
    ## the aforementioned requests, the new InspectLoop request
    ## can become ripe before they are completed, and we reenter
    ## this proc without servicing background processes.
    after 20 InspectLoop $solver
}

# Initialize display graph
set graph_frame [eval frame .graph_frame  $frame_opts]
set graph_select_frame [frame $graph_frame.select_frame]
set graph_holding_frame [frame $graph_frame.holding -bd 0]
## The invidual graphs are packed into the "holding frame" rather than
## directly into $graph_frame.  This way $graph_holding_frame" can
## retain the graph widget size and prevent unnecessary configure
## (and subsequent graph rescaling) events when one graph is turned
## off and another on.

pack  $graph_select_frame -side bottom -fill x -expand 0
pack $graph_holding_frame -side bottom -fill both -expand 1


proc MakeGraph { basewin labelx labely args } {
    Ow_GraphWin New newgraph $basewin -curve_width 3 \
            -xlabel $labelx -ylabel $labely \
            -outer_frame_options "-bd 0"
    foreach curve $args { $newgraph NewCurve $curve }
    return $newgraph
}

set graphTi [MakeGraph $graph_holding_frame "Iteration" "|m x h|" "|m x h|"]
set graphMi [MakeGraph $graph_holding_frame "Iteration" "M/Ms" \
	"Mx" "My" "Mz"]
set graphEi [MakeGraph $graph_holding_frame "Iteration" "Energy" \
	"Total\nEnergy" "Exchange\nEnergy" "Anisotropy\nEnergy" \
	"Demag\nEnergy" "Zeeman\nEnergy"]
set graphAngi [MakeGraph $graph_holding_frame "Iteration" "Max Angle" \
	"Max\nAngle"]
set graphTb [MakeGraph $graph_holding_frame $Bdir_label "|m x h|" "|m x h|"]
set graphMb [MakeGraph $graph_holding_frame $Bdir_label "M/Ms" \
	"Mx" "My" "Mz"]
set graphEb [MakeGraph $graph_holding_frame $Bdir_label "Energy" \
	"Total\nEnergy" "Exchange\nEnergy" "Anisotropy\nEnergy" \
	"Demag\nEnergy" "Zeeman\nEnergy"]
set graphAngb [MakeGraph $graph_holding_frame $Bdir_label "Max Angle" \
	"Max\nAngle"]
set graph_select_abscissa [eval frame $graph_select_frame.abscissa $frame_opts]
set graph_select_ordinate [eval frame $graph_select_frame.ordinate $frame_opts]
pack $graph_select_abscissa $graph_select_ordinate \
	-side left -fill both -expand 1

set abscissa_select 0
set iter_rb [radiobutton $graph_select_abscissa.iter \
	-command SelectGraph \
	-variable abscissa_select -text Iter -value 0]
set bapp_rb [radiobutton $graph_select_abscissa.bapp \
	-command SelectGraph \
	-variable abscissa_select -text "B app" -value 1]
pack $iter_rb $bapp_rb -side left -fill both -expand 1

set ordinate_select 0
set mxh_rb [radiobutton $graph_select_ordinate.mxh \
	-command SelectGraph \
	-variable ordinate_select -text "|m x h|" -value 0]
set mag_rb [radiobutton $graph_select_ordinate.mag \
	-command SelectGraph \
	-variable ordinate_select -text "Mag" -value 1]
set energy_rb [radiobutton $graph_select_ordinate.energy \
	-command SelectGraph \
	-variable ordinate_select -text "Energy" -value 2]
set maxang_rb [radiobutton $graph_select_ordinate.maxang \
	-command SelectGraph \
	-variable ordinate_select -text "Max Ang" -value 3]

pack $mxh_rb $mag_rb $energy_rb $maxang_rb -side left -fill both -expand 1
set graph_array(0,0) $graphTi    ;   set graph_array(1,0) $graphTb
set graph_array(0,1) $graphMi	 ;   set graph_array(1,1) $graphMb
set graph_array(0,2) $graphEi	 ;   set graph_array(1,2) $graphEb
set graph_array(0,3) $graphAngi	 ;   set graph_array(1,3) $graphAngb

proc SelectGraph { } {
    global current_graph abscissa_select ordinate_select graph_array
    set new_graph $graph_array($abscissa_select,$ordinate_select)
    if {[info exists current_graph]} {
	if {[string compare $current_graph $new_graph]==0} { return }
	pack forget [$current_graph Cget -winpath]
    }
    set current_graph $new_graph
    pack [$current_graph Cget -winpath] -side bottom -fill both -expand 1
    UpdateCurrentGraph
}

SelectGraph   ;# Initial default display depends on the values of
## abscissa_select ordinate_select set above

# Field control block
set btn_frame [frame .btn_frame]
set field_frame [eval frame .field_frame $frame_opts]
set field_value [label $field_frame.value -relief groove \
	-text [$solver GetNomAppliedField $Bfield_multiplier \
	       "(%.1f %.1f %.1f)"]]
set field_label [label $field_frame.label -relief groove -text "Field (mT)"]
set field_minus [button $field_frame.minus -text "-" \
	-command "FieldAdjust -1"]
set field_plus [button $field_frame.plus -text "+" \
	-command "FieldAdjust 1"]
pack $field_value -ipadx 2 -ipady 4 -side top -fill both -expand 0
pack $field_minus -side left -fill both -expand 0
pack $field_label -ipadx 2 -side left -fill both -expand 0
pack $field_plus  -side left -fill both -expand 0

proc FieldAdjust { step } {
    global solver field_value Bfield_multiplier

    if {$step != 0} {
	UpdateOutput bapp   ;# Assume user wants current state
	## recorded.  Note that UpdateOutput has a built-in check
	## to assure no iterate gets outputed more than once to
	## each output channel.
	$solver IncrementFieldStep $step
    }

    $field_value configure \
	    -text [$solver GetNomAppliedField $Bfield_multiplier \
	           "(%.1f %.1f %.1f)"]
}

# Other control buttons
set ctrl_frame [eval frame .ctrl_frame $frame_opts]
set btn_reset [button $ctrl_frame.b0 -text "Reset Solver" \
	-command "ChangeSolverState Pause ; after idle $solver Reset"]
set btn_run [button $ctrl_frame.b1 -text "Run" \
	-command "ChangeSolverState Run"]
set btn_relax [button $ctrl_frame.b2 -text "Relax" \
	-command "ChangeSolverState Relax"]
set btn_pause [button $ctrl_frame.b3 -text "Pause" \
	-command "ChangeSolverState Pause"]
set btn_print_mag [button $ctrl_frame.bmag -text "Write Mag" \
	-command "$solver WriteMagFile"]
set btn_print_field [button $ctrl_frame.bfield -text "Write Field" \
	-command "$solver WriteFieldFile"]
set btn_print_labels [button $ctrl_frame.b4 -text "Print Labels" \
	-command PrintLabels]
set btn_print_info [button $ctrl_frame.b5 -text "Print Info" \
	-command "PrintState $solver"]
set btn_quit [button $ctrl_frame.b6 -text "Quit" \
	-command "catch \{$solver Delete\} ; exit"]
pack $btn_reset $btn_run $btn_relax $btn_pause \
	$btn_print_mag $btn_print_field \
	$btn_print_labels $btn_print_info \
	$btn_quit -side top -fill both -expand 1

proc ChangeSolverState { newstate } {
    # newstate should be one of Run, Relax, or Pause.
    #
    # NOTE 1: Restarting the ODE solver after a pause resets the
    #  internal solver state, i.e., you lose previous magnetization
    #  and field state info used in by the predictor-corrector
    #  algorithm.  As a result, interactive pausing may cause the
    #  solver results to be *slightly* different.
    #
    # NOTE 2: This proc is a little tricky because it is reentrant.

    global btn_run btn_relax btn_pause solver oldstate

    if {[info exists oldstate] && [string compare $oldstate $newstate]==0} {
	return  ;# No change
    }

    # Pause solver if necessary
    if {[info exists oldstate] && ![string match "Pause" $oldstate]} {
	$solver ChangeSolverState Pause
	after 20 ChangeSolverState $newstate
	return
	## Let solver process pause request
    }

    # Go to new state
    set oldstate $newstate
    if { [string match "Run" $newstate] } {
        $btn_pause configure -relief raised
	$btn_run configure -relief sunken
	update
	InspectLoop $solver
	$solver HystLoop "UpdateOutput bapp" "FieldAdjust 0"
        ## It may be a long time before the above call returns
	$btn_run configure -relief raised
    } elseif { [string match "Relax" $newstate] } {
        $btn_pause configure -relief raised
	$btn_relax configure -relief sunken
	update
	InspectLoop $solver
	$solver RelaxState "UpdateOutput bapp"
        ## It may be a long time before the above call returns
	$btn_relax configure -relief raised
    }

    # Upon return from the HystLoop or RelaxState calls,
    # or in the event of a Pause solver request, go
    # into pause mode
    InspectLoop $solver ;# Process any pending output
    after cancel InspectLoop $solver ;# No mas!
    $btn_pause configure -relief sunken
    set oldstate "Pause"
    update
}

pack $field_frame -in $btn_frame -side top -fill both -expand 0
pack $ctrl_frame -in $btn_frame -side top -fill both -expand 1
pack $btn_frame -side left -fill y
pack $graph_frame -side top -fill both -expand 1
