# FILE: thread.tcl
#
# An object which plays the role of a thread within an interpreter.
#
# The Tcl event loop must be running in order for non-blocking socket
# communications to work.
#

Oc_Class Net_Thread {

    # Index of thread objects by hostname,pid
    array common index {}

    # Name of host where the account this object represents resides
    const public variable hostname = localhost

    # Name of account this object represents
    const public variable accountname

    # Process ID of the thread
    const public variable pid

    # The "alias" of this thread
    const public variable alias

    # The account under which this thread is running
    private variable account

    # Our connection to the actual running thread
    private variable connection

    # Port on host on which the thread server listens
    private variable port
    
    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler delete
        }
        set logHandler [em_eventHandler new $log delete "$this Delete"]
        if {$imadelog} {
            $oldlog Delete
        }
        set oldlog $log
    }
    private variable oldlog = Oc_Nop
    private variable logHandler
    private variable imadelog = 0

    private variable msgHandler

    # Send id
    private variable id = 0

    # Reply message
    private variable reply

    # Is object ready to relay messages to and from the thread?
    private variable ready = 0

    # Event handling support
    private variable bindtags
    private variable myEventHandlers = {}

    private variable protocolname
    private variable protocolversion

    method Protocol {} {
        return $protocolname
    }

    public variable gui = 0 {
        if {$gui} {   
            if {[info exists myWidget]} {
                return
            }
            package require Tk
            Net_ThreadGui New myWidget -thread $this
        } else {
            if {![info exists myWidget]} {
                return
            }
            $myWidget Delete
            unset myWidget
        }
        set btnvar $gui
    }    
    private variable myWidget
    private variable hasGui = 0
    private variable btnvar

    method HasGui {} {
        return $hasGui
    }
 
    method ToggleGui { } {
        $this Configure -gui $btnvar
    }

    # Constructor is a non-blocking operation.  
    Constructor { args } {
        eval $this Configure $args
        if {![info exists accountname]} {
            set accountname [Net_Account DefaultAccountName]
        }
        if {![info exists pid]} {
            error "Thread must have pid!"
        }
        # Enforce only one object per account
        if {![catch {set index($hostname,$pid)} existingthread]} {
            unset hostname
            unset pid
            $this Delete
            array set argarray $args
            catch {unset argarray(-hostname)}
            catch {unset argarray(-accountname)}
            catch {unset argarray(-pid)}
            catch {unset argarray(-alias)}
            eval $existingthread Configure [array get argarray]
            return $existingthread
        }
        array set index [list $hostname,$pid $this]
        if {![info exists log]} {
            # Provide a default log
            $this Configure -log [em_log New _ \
                    -title "Thread $hostname:$pid Log"]
            set imadelog 1
        }
        # Enable event generation
        set bindtags [list $this $class]
        if {[catch {
                Net_Account New account -hostname $hostname -log $log \
                -accountname $accountname} msg]} {
            error "Can't construct account object.\n$msg"
        } else {
            lappend myEventHandlers [em_eventHandler new $account delete \
                    [list $this Delete]]
            if {[$account Ready]} {
                $this LookupThreadPort
            } else {
                lappend myEventHandlers [em_eventHandler new $account Ready \
                        [list $this LookupThreadPort]]
            }
        }
    }

    method LookupThreadPort {} {
        set qid [$account Send lookup $pid]
        Oc_EventHandler New _ $account Reply$qid [list $this \
                GetThreadPort] -oneshot 1 -groups [list $this $this-gtp]
        Oc_EventHandler New _ $account Reply$qid [list $this \
                NoAnswerFromAccount] -oneshot 1 -groups [list $this $this-gtp]
    }

    method GetThreadPort {} {
        Oc_EventHandler DeleteGroup $this-gtp
        set reply [$account Get]
        if {[llength $reply] < 2} {
            $log Append "Bad reply from account $accountname: $reply"
            error "Bad reply from account $accountname: $reply"
        }
        switch -- [lindex $reply 0] {
            0 {
                set port [lindex $reply 1]
                $this Connect
            }
            default {
                # Account returns an error - assume that means lookup failed
                set msg "No thread $pid registered with account\
                        [$account Cget -accountname]:\n\t[lrange $reply \
                        1 end]"
                $this Delete
                global errorInfo
                set errorInfo "No stack available"
                Oc_Log Log $msg warning $class
            }
        }
    }

    method NoAnswerFromAccount {} {
        Oc_EventHandler DeleteGroup $this-gtp
        # Host not responding -- give up.
        set msg "No reply from account [$account Cget -accountname]"
        $this Delete
        global errorInfo
        set errorInfo "No stack available"
        Oc_Log Log $msg warning $class
    }

    method Connect {} {
        if {[catch {
                Net_Connection New connection -hostname $hostname -port $port \
                -log $log}]} {
            $this Delete
        } else {
            lappend myEventHandlers [em_eventHandler new $connection delete \
                    [list $this Delete]]
            lappend myEventHandlers [em_eventHandler new $connection Ready \
                    [list $this ConnectionReady]]
        }
    }

    method ConnectionReady {} {
        set msgHandler [em_eventHandler new $connection Readable \
                [list $this VerifyOOMMFThread]]
    }

    method VerifyOOMMFThread {} {
        $msgHandler delete
        unset msgHandler
        if {![regexp {OOMMF ([^ ]+) protocol ([0-9.]+)} [$connection Get] \
                match name version]} {
            # Not an OOMMF thread
            set msg "No OOMMF thread server answering on $hostname:$port"
            $this Delete
            global errorInfo
            set errorInfo "No stack available"
            Oc_Log Log $msg warning $class
            return
        }
        set protocolname $name
        set protocolversion $version
        $log Append "OOMMF thread $hostname:$pid ready"
        set msgHandler [em_eventHandler new $connection Readable \
                [list $this Hear]]
        set ready 1
        incr id
        set qid [$connection Query HasGui]
        Oc_EventHandler New _ $connection Reply$qid [list $this \
                ReceiveHasGui] -oneshot 1 -groups [list $this $this-gui]
        Oc_EventHandler New _ $connection Timeout$qid [list $this \
                Timeout gui] -oneshot 1 -groups [list $this $this-gui]
    }

    method ReceiveHasGui {} {
        Oc_EventHandler DeleteGroup $this-gui
        set reply [$connection Get] 
        if {[lindex $reply 0]} {
            set emsg "Error report from $alias<$pid>:\n\t[lrange $reply \
                    1 end]\n\nIn reply to query:\n\tHasGui"
            error $emsg $emsg
        } else {
            set hasGui [lindex $reply 1]
        }
        em_eventHandler Generate Ready $bindtags
    }

    # Handle 'tell' messages from connection
    method Hear {} {
        set reply [$connection Get]
        em_eventHandler Generate Readable $bindtags
    }

    method Receive { rid } {
        Oc_EventHandler DeleteGroup $this-$rid
        set reply [$connection Get] 
        Oc_EventHandler Generate $this Reply$rid
    }

    method Get {} {
        return $reply
    }

    # Send a command to the thread server
    method Send { args } {
        if {!$ready} {
            error "thread $hostname:$pid not ready"
        }
        incr id
        set qid [eval $connection Query $args]
        Oc_EventHandler New _ $connection Reply$qid [list $this \
                Receive $id] -oneshot 1 -groups [list $this $this-$id]
        Oc_EventHandler New _ $connection Timeout$qid [list $this \
                Timeout $id] -oneshot 1 -groups [list $this $this-$id]
        return $id
    }

    method Timeout { rid } {
        Oc_EventHandler DeleteGroup $this-$rid
        Oc_EventHandler Generate $this Timeout$rid
    }

    method Ready {} {
        return $ready
    }

    Destructor {
        set ready 0
        if {[info exists bindtags]} {
            em_eventHandler Generate delete $bindtags
        }
        # delete handlers before deleting event-generating members
        Oc_EventHandler DeleteGroup $this
        foreach handler $myEventHandlers {
            $handler delete
        }
        if {[info exists msgHandler]} {
            $msgHandler delete
        }
        if {[info exists connection]} {
            $connection Delete
        }
        if {[info exists log]} {
            $log Append "Deleting $this ..."
            $this Configure -log Oc_Nop
        }
        if {[info exists myWidget]} {
            $this Configure -gui 0
        }
        if {[info exists hostname] && [info exists pid]} {
            unset index($hostname,$pid)
        }
    }

}
